module Main( main ) where
import Version
import PPrint
import HParse 
import Syntax
import Infer
import Types 
import Utils
import Constants
import Expressions
import Substitutions
import SubExpressions
import Matching
import Names
import List
import Calculations
import Rewrite
import Laws
import System
import Html (renderHtml,prettyHtml)
import State
import UU_Scanner
import UU_Parsing
import UU_Parsing_ext
import GetOpt
import IO

type Path = String

data ProgState = ProgState { 
                   path :: Path, 
                   env :: Env,
                   theor :: Theor,
                   echo :: Bool,
                   debugfile :: String,
                   matcher :: Matcher,
                   width :: Int,
                   theormap :: TheorMap,
                   sourcefile :: SourceFile,
                   trafomap :: TrafoApplyMap
                 }

initstate = ProgState { 
              path = "",
              env = error "no program loaded",
              theor = undefined,
              echo = False,
              debugfile = "",
              matcher = fastmatcher,
              width = 80::Int,
              theormap = undefined,
              sourcefile = error "no program loaded",
              trafomap = undefined
            }

type Program = StT ProgState

getSourceFile :: Monad m => Program m SourceFile
getSourceFile = do st <- readSt
                   return (sourcefile st)

loadSourceFile :: Monad m => 
                    (SourceFile -> Program m SourceFileInfo)
                              -> SourceFile -> Program m ()
loadSourceFile tc sourcefile 
         = do st <- readSt
              sourcefileinfo <- tc sourcefile
              writeSt (st { theormap = sfitheormap sourcefileinfo, 
                            trafomap = sfitrafomap sourcefileinfo,
                            env = sfienv sourcefileinfo, 
                            sourcefile = sourcefile })
 
getTheoryMap :: Monad m => Program m TheorMap
getTheoryMap = do st <- readSt
                  return (theormap st)

getTrafoApplyMap :: Monad m => Program m TrafoApplyMap
getTrafoApplyMap = do st <- readSt
                      return (trafomap st)

{-
setTheoryMap :: Monad m => TheorMap -> Program m ()
setTheoryMap tm = do st <- readSt
                     writeSt (st { theormap = tm })
-}

getProgram :: Monad m => Program m Env
getProgram = do st <- readSt 
                return (env st)

{-
setProgram :: Monad m => Env -> Program m ()
setProgram p = do st <- readSt
                  writeSt (st { env = p })
-}

getTheory :: Monad m => Program m Theor
getTheory = do st <- readSt
               return (theor st)

setTheory :: Monad m => Theor -> Program m ()
setTheory t = do st <- readSt
                 writeSt (st { theor = t })

getPath :: Monad m => Program m Path
getPath = do st <- readSt
             return (path st)

setPath :: Monad m => Path -> Program m ()
setPath d = do st <- readSt
               writeSt (st { path = pathify d })
                  where pathify d = if (not (null d) && last d /= '/') 
                                    then d++"/" else d

getEcho :: Monad m => Program m Bool
getEcho = do st <- readSt
	     return (echo st)

setEcho :: Monad m => Bool -> Program m ()
setEcho e = do st <- readSt
	       writeSt (st { echo = e })

getDebugFile :: Monad m => Program m String
getDebugFile = do st <- readSt
		  return (debugfile st)

setDebugFile :: Monad m => String -> Program m ()
setDebugFile s = do st <- readSt
		    writeSt (st { debugfile = s })

getMatcher :: Monad m => Program m Matcher
getMatcher = do st <- readSt
                return (matcher st)

setMatcher :: Monad m => Matcher -> Program m ()
setMatcher m = do st <- readSt
		  writeSt (st { matcher = m })

getWidth :: Monad m => Program m Int
getWidth = do st <- readSt
              return (width st)

setWidth :: Monad m => Int -> Program m ()
setWidth w = do st <- readSt
		writeSt (st { width = w })

options :: [OptDescr Command]
options = ghcioptions ++ 
 [ Option [] ["program"]   (ReqArg LoadProgram "filename") 
                                              "program file to load"
 , Option [] ["theory"]    (ReqArg LoadTheory "filename")       
                                              "theory file to load"
 , Option [] ["path"]      (ReqArg SetPath "path")  "path to use"
 , Option [] ["run"]       (ReqArg RunScript "script") "script to run"
 , Option [] ["debug"]     (ReqArg DebugFile "file") 
                                 "HTML debugging output for rewriting process"
 , Option [] ["width"]     (ReqArg (SetWidth . stringToInt) "width") 
                                 "pretty printing width"
 , Option [] ["transform"] (ReqArg ApplyTransform "transformation")
                                 "transformation to apply"
 ]

skipshortoption c = Option [c] [] (NoArg Skip) ""
skiplongoption s = Option [] [s] (NoArg Skip) ""

ghcioptions :: [OptDescr Command]
ghcioptions = map skipshortoption "rlibghcBus/" ++ 
              map skiplongoption ["interactive","5.00.2"]

main = 
   evalSt initstate $
    do 
       -- Can't do this because hugs doesn't support it - instead use hFlush
       -- after displaying "Mag> " prompt.
       -- hSetBuffering stdout NoBuffering
       args <- liftSt getArgs
       let (commands, rest, errors) = getOpt RequireOrder options args
       if null errors then
         do sequence $ map runCommand commands
            if null rest then do liftSt $ putStrLn ("MAG version " 
                                                     ++ versionstr
                                                     ++ " (built on " 
                                                     ++ datestr ++ ")")
                                 commandLoop stdin
             else do let rest' = head rest:map (\s->"\""++s++"\"") (tail rest)
                     command <- liftSt $ parseCommand (sepby " " rest')
		     runCommand command
        else liftSt $ putStr $ sepby "\n" errors

data Command = SetPath String 
	     | SetWidth Int
	     | DebugFile String
             | RunScript String
	     | SetMatcher Matcher
	     | ShowTheory
             | ShowProgram
             | LoadProgram String 
             | LoadTheory String 
             | ChooseTheory String
             | DoRewrite String
             | ApplyTransform String
             | SimpleMatch String String 
             | OneStepMatch String String
             | TwoStepMatch String String
             | Quit 
	     | Skip

commandstxt = ["loadprogram","loadtheory","rewrite","applytransform",
               "simplematch",
               "onestepmatch","twostepmatch","quit","setpath","runscript",
	       "skip","debugfile","setmatcher","setwidth",
	       "ref","fast",
	       "compare","onesteponly","check","trace",
	       "showtheory","showprogram",
               "choosetheory"]

commandScan x = scanner commandstxt [] "()" "" x 1

pLoadProgram = LoadProgram <$-> pKey "loadprogram" <*> pString
pLoadTheory = LoadTheory <$-> pKey "loadtheory" <*> pString
pChooseTheory = ChooseTheory <$-> pKey "choosetheory" <*> pString
pRewrite = DoRewrite <$-> pKey "rewrite" <*> pString
pApplyTransform = ApplyTransform <$-> pKey "applytransform" <*> pString
pSimpleMatch = SimpleMatch <$-> pKey "simplematch" <*> pString <*> pString
pOneStepMatch = OneStepMatch <$-> pKey "onestepmatch" <*> pString <*> pString
pTwoStepMatch = TwoStepMatch <$-> pKey "twostepmatch" <*> pString <*> pString

pQuit = Quit <$-> pKey "quit"
pSetPath = SetPath <$-> pKey "setpath" <*> pString
pRunScript = RunScript <$-> pKey "runscript" <*> pString
pSkip = Skip <$-> pKey "skip"
pDebugFile = DebugFile <$-> pKey "debugfile" <*> pString
pSetWidth = (SetWidth . stringToInt) <$-> pKey "setwidth" <*> pInteger
pSetMatcher = SetMatcher <$-> pKey "setmatcher" <*> pMatcher
pShowTheory = ShowTheory <$-> pKey "showtheory" 
pShowProgram = ShowProgram <$-> pKey "showprogram"

pRefMatcher = referencematcher <$-> pKey "ref"
pFastMatcher = fastmatcher <$-> pKey "fast"

pCompareMatcher = comparematchers <$-> pKey "compare" <*> pMatcher <*> pMatcher
pOneStepOnly = forceonestep <$-> pKey "onesteponly" <*> pMatcher
pCheckMatcher = checkmatcher <$-> pKey "check" <*> pMatcher
pTraceMatcher = tracematcher <$-> pKey "trace" <*> pMatcher

pMatcher = pRefMatcher <|> pFastMatcher 
           <|> pOneStepOnly <|> pCompareMatcher 
           <|> pCheckMatcher <|> pTraceMatcher
           <|> pparens pMatcher

pCommand = pLoadProgram <|> pLoadTheory <|> pChooseTheory <|> 
               pRewrite <|> pApplyTransform <|>
               pSimpleMatch <|> pOneStepMatch <|> pTwoStepMatch <|>
               pQuit <|> pSetPath <|> pRunScript <|> pDebugFile <|>
	       pSetMatcher <|> pShowTheory <|> pShowProgram <|>
               pSetWidth <|>
	       pSkip <|> succeed Skip

parseCommand :: String -> IO Command
parseCommand commandString =
   do let (command,reset,cfree_errors) 
           = invokeErr pCommand (commandScan commandString) []
      putStr cfree_errors
      return command 

handleEOF error = if isEOFError error then liftSt $ putStr "\n"
                                      else liftSt $ print error

commandLoop :: Handle -> Program IO ()
commandLoop h = 
 flip catchSt handleEOF $
       do liftSt $ putStr "Mag> "
          liftSt $ hFlush stdout
          commandString <- liftSt $ lineBreaks $ hGetLine h
          echo <- getEcho
          liftSt $ putStrLn (if echo then commandString else "")
	  command <- liftSt $ parseCommand commandString
          case command of 
                 Quit -> return ()
                 _    -> do let loop = commandLoop h
	                    catchSt (runCommand command) showerror
                            loop
              where showerror error = do liftSt $ print error


lineBreaks :: IO String -> IO String
lineBreaks getline = do str <- getline
			if (not (null str) && last str == '\\') 
			   then do rest <- lineBreaks getline
				   return $ init str ++ rest
			   else return str

dorewrite :: Theor -> Exp -> Program IO Calculation
dorewrite th exp = do m <- getMatcher
                      w <- getWidth
                      let calc = calculate m th (exp,[])
                      liftSt $ putStr $ prettyCalc w 0 calc
                      liftSt $ putStr "\n\n"
                      debugfile <- getDebugFile
                      if not (null debugfile) 
                       then do liftSt $ writeFile debugfile 
                                  $ renderHtml $ htmlCalc calc
                               return calc
                       else return calc
                  
-- probably should thaw stuff properly
newdefs :: Theor -> (String,Type) -> Program IO [Def]
newdefs th (name,typ) 
  = do calc <- dorewrite th (Con (Constr name) typ)
       return [Fun (Var name typ) (unfreeze $ conclusion calc)]
   where unfreeze (Lam p b t) = Lam p (unfreeze b) t
         unfreeze (Ap f a t) = Ap (unfreeze f) (unfreeze a) t
         unfreeze (Var x t) = Var x t
         unfreeze (Con (Constr x) t) = Var x t
         unfreeze (Con c t) = Con c t

runCommand :: Command -> Program IO ()
runCommand (LoadProgram filename) 
            = do d <- getPath
                 input <- liftSt $ readFile (d++filename++".p")
                 sourcefile <- liftSt $ parse input
                 loadSourceFile (liftSt . typecheck) sourcefile

runCommand (LoadTheory filename) = do d <- getPath
                                      p <- getProgram
                                      th <- liftSt $ gettheo p (d++filename)
                                      setTheory th

runCommand (ChooseTheory transform) 
       = do tm <- getTheoryMap
            case tm transform of
              Just theor -> setTheory theor
              Nothing -> liftSt $ putStrLn "transform not found"

runCommand (ApplyTransform transform)
       = do tam <- getTrafoApplyMap
            case tam transform of
              Nothing -> liftSt $ putStrLn "transform not found"
              Just ta -> do newsourcefile <- applytrafo ta newdefs
                            loadSourceFile (liftSt . typecheck) newsourcefile

runCommand (DoRewrite expStr) = do p <- getProgram
                                   th <- getTheory
                                   exp <- liftSt $ getexp expStr p
                                   _ <- dorewrite th exp
                                   return ()

runCommand (SimpleMatch patternStr termStr) =
    do m <- getMatcher
       (pattern,term) <- parseEqn patternStr termStr
       liftSt $ putStr $ prettySs $ simplematch m pattern term

runCommand (OneStepMatch patternStr termStr) = 
    do m <- getMatcher
       (pattern,term) <- parseEqn patternStr termStr
       let ss = evalName 
                 (suppex (allvars pattern `union` allvars term) [] supply)
                                   (onestepmatch m pattern (clos term))
       liftSt $ putStr $ prettySs ss


runCommand (TwoStepMatch patternStr termStr) = 
    do m <- getMatcher
       (pattern,term) <- parseEqn patternStr termStr
       let ss = evalName 
                 (suppex (allvars pattern `union` allvars term) [] supply)
                                   (twostepmatch m pattern (clos term))
       liftSt $ putStr $ prettySs ss

runCommand (SetPath dir) = setPath dir

runCommand (SetMatcher matcher) = setMatcher matcher

runCommand (SetWidth width) = setWidth width

runCommand ShowTheory = do theory <- getTheory
                           liftSt $ putStrLn $ prettyTheo theory

runCommand ShowProgram = do sourcefile <- getSourceFile
                            w <- getWidth
                            liftSt $ putStrLn $ wdisplay w 
                                              $ ppSourceFile sourcefile

runCommand (RunScript script) 
      = do dir <- getPath
	   oldEcho <- getEcho
	   setEcho True
           h <- liftSt $ openFile (dir++script) ReadMode
	   pos <- liftSt $ hGetPosn h
           flip catchSt handleEOF $
              do firstline <- liftSt $ hGetLine h
                 if (take 2 firstline /= "#!") 
                    then liftSt $ hSetPosn pos 
                    else return ()
                 commandLoop h
           liftSt $ hClose h
	   setEcho oldEcho

runCommand (DebugFile file) = setDebugFile file

runCommand Skip = return ()

parseEqn :: String -> String -> Program IO (Exp,Exp)

parseEqn e1Str e2Str = 
   do p <- getProgram
      e1 <- liftSt $ getexp e1Str p
      e2 <- liftSt $ getexp e2Str p
      [(_,[],(e1',e2'))] <- liftSt $ typetheo p (theo e1 e2)
      return (e1',e2')

oldmag = do putStr "MAG calculator\n"
            sourcefileinfo <- promptprog
            th    <- prompttheo (sfienv sourcefileinfo)
	    let loop = do putStr "Type expression: "
                          x <- getLine
			  if not (null x) then
				  (do xt <- getexp x (sfienv sourcefileinfo)
				      dispCalc th xt
            			      loop
				  )
                           else return ()
            loop

dispCalc th xt = let calc = calculate fastmatcher th (xt,[]) in
                  do putStr $ prettyCalc 80 0 $ calc
                     putStr "\n\n"

-------------------
-- testing programs

progType    = prog "program types" (showEnv.sfienv)
progPretty  = prog "program pretty printer" (prettyStr.sfiprog)

----------------------
-- testing expressions

expType     = progexp "expression type" allTypes
expBeta     = progexp "beta reduction"  (prettyExp . betareduce)
expEta      = progexp "eta contraction" (prettyExp . etaContract)
expShow     = progexp "display abstract syntax (no types)" showExp
expFullShow = progexp "display abstract syntax" show
                 
showExp (Var x _) = "Var "++x
showExp (Con c _) = "Con ("++show c++")"
showExp (Lam p b _) = "Lam ("++showExp p++") ("++showExp b++")"
showExp (Ap f e _) = "Ap ("++showExp f++") ("++showExp e++")"

-------------------
-- testing theories

theoPretty = theory "theory scrub and pretty printer" (prettyTheo . scrubth)

--------------------
-- testing equations

eqnEqual = eqn "equality" (\ e1 e2 -> show (e1 == e2))
eqnMatch = eqn "match" (\ e1 e2 -> prettySs (simplematch fastmatcher e1 e2))
eqnHighMatch 
   = eqn "high match" 
         (\ e1 e2 -> 
	     let ss = evalName 
	                 (suppex (allvars e1 `union` allvars e2) [] supply)
                                   (twostepmatch fastmatcher e1 (clos e2))
             in prettySs ss)

----------------------------------------------
-- reading programs, expressions and equations

promptprog = do putStr "Type program name <filename>.p: "
                filename  <- getLine
                getprog filename
                
getprog filename = do input     <- readFile (filename++".p")
                      prog      <- parse input
                      typecheck prog

prompttheo e = do putStr "Type theory name <filename>.eq: "
                  filename  <- getLine
                  gettheo e filename

gettheo e filename = do input     <- readFile (filename++".eq")
                        th        <- ptheory input
                        typetheo e th

promptexp s e = do putStr ("Type "++ s ++ ": ")
                   x  <- getLine
                   getexp x e 
                 
getexp x e = do x' <- parseExp x
                typecheckExp e x'


prompteqn e = do p <- promptexp "pattern" e
                 t <- promptexp "term" e
                 [(_,[],(e1,e2))] <- typetheo e (theo p t)
                 return (e1,e2)

progexp s f = do putStr (s ++ "\n")
                 sfi <- promptprog
                 x <- promptexp "expression" (sfienv sfi)
                 putStr (f x)

prog s f = do putStr (s++ "\n")
              ep <- promptprog
              putStr (f ep)

theory s f = do putStr (s++ "\n")
                sfi <- promptprog
                t <- prompttheo (sfienv sfi)
                putStr (f t)

eqn s f = do putStr (s++"\n")
             sfi <- promptprog
             (e1,e2) <- prompteqn (sfienv sfi)
             putStr (f e1 e2)

----------------------------------
-- constructing singleton theories
             
typecheckExp e x = do [(_,[],(y,_))] <- typetheo e (theo x (Con Error errType))
                      return y 

theo e1 e2 = Theory (CeqTheo (Ceq "" NoEq (Eqn e1 e2) 0) NoTheo)

--------------------------------------
-- special pretty printing for Main.hs

showEnv :: Env -> String
showEnv = sepby '\n' . map showDecl
          where sepby c  = foldr1 step
                           where step x y = x ++ [c] ++ y

showDecl :: (String,Type) -> String
showDecl (x,t) = x ++ " :: " ++ prettyType t



